﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace Dcapture
{
    public class GraphDrawer
    {
        Bitmap bitmapWhite = null;
        Bitmap bitmapPrevious = null;
        uint[,] zeroMatrix = null;
        Color[] colorScale = new Color[256];
        
        public GraphDrawer(int bitmapSizeInBits, MatrixCSV mainMatrix, MatrixCSV shadowMatrix, PhotonNeutronSeparation photonNeutronSeparation)
        {
            this.mainMatrix = mainMatrix;
            this.shadowMatrix = shadowMatrix;
            this.photonNeutronSeparation = photonNeutronSeparation;
            
            //define and clear white bitmap and zeroMatrix
            this.bitmapWhite = new Bitmap((int)Math.Pow(2, bitmapSizeInBits), (int)Math.Pow(2, bitmapSizeInBits));
            this.zeroMatrix = new uint[this.bitmapWhite.Width, this.bitmapWhite.Height];
            for (int i = 0; i < bitmapWhite.Height; i++)
            {
                for (int j = 0; j < bitmapWhite.Width; j++)
                {
                    bitmapWhite.SetPixel(i, j, Color.White);
                    zeroMatrix[i, j] = 0;
                }
            }

            //define color scale
            for (int i = 0; i < 256; i++)
            {
                colorScale[i] = Dcapture.Properties.Resources.black_blue_red_yellow.GetPixel(0, 255 - i);
            }   
        }

        private MatrixCSV mainMatrix = null;
        private MatrixCSV shadowMatrix = null;
        private PhotonNeutronSeparation photonNeutronSeparation = null;

        private uint maxNumSamplesInGraph = 0;
        public uint MaxNumSamplesInGraph
        {
            get { return maxNumSamplesInGraph; }
        }

        public struct GrawGraphArguments
        { 
            public int positionDiscrParam;
            public int discrParam_endIndex;
            public int discrParam_divForActualZoom;
            public int discrParam_zoomBitmap;
            public int positionEnergy;
            public int energy_endIndex;
            public int energy_divForActualZoom;
            public int energy_zoomBitmap;
            public GraphControl.ShadowMatrixDrawingStyle shadowMatrixDrawingStyle;
            public int colorSaturation;
            public bool redrawOnlySectionPreview;
        }

        /// <summary>
        /// compute coordinate in graph from coordinate in matrix
        /// </summary>
        /// <param name="coordInMatrix"></param>
        /// <returns>return -1 if value is out of the graph</returns>
        private int ComputeEnergyCoords(int coordInMatrix, GrawGraphArguments args)
        {
            int coordInGraph = -1;
            if (coordInMatrix >= args.positionEnergy)
            {
                coordInGraph = (coordInMatrix - args.positionEnergy) / args.energy_divForActualZoom;
            }

            return coordInGraph;
        }

        public Image DrawGraph(GrawGraphArguments args)
        {
            //initiate matrix same size as bitmap
            uint[,] mainM = (uint[,])zeroMatrix.Clone();
            uint[,] shadowM = (uint[,])zeroMatrix.Clone();
            
            //check indexes (because of vScrollBar gives stupid valuse - one more grater than I expected)
            if (args.positionDiscrParam + args.discrParam_endIndex >= MatrixCSV.DiscrParamSizeMatrix)
            {
                Console.WriteLine("Index will be grater than matrix dimension");
                args.positionDiscrParam = MatrixCSV.DiscrParamSizeMatrix - args.discrParam_endIndex;
            }
            if (args.positionEnergy + args.energy_endIndex >= MatrixCSV.EnergySizeMatrix)
            {
                Console.WriteLine("Index will be grater than matrix dimension");
                args.positionEnergy = MatrixCSV.EnergySizeMatrix - args.energy_endIndex;
            }
            
            //make one matrix
            uint maxMainM = 1;
            uint maxShadowM = 1;
            if (!args.redrawOnlySectionPreview)
            {
                for (int iDiscrParam = 0; iDiscrParam < args.discrParam_endIndex; iDiscrParam++)
                {
                    for (int iEnergy = 0; iEnergy < args.energy_endIndex; iEnergy++)
                    {
                        //mainMatrix
                        uint origValue = mainMatrix[args.positionDiscrParam + iDiscrParam, args.positionEnergy + iEnergy];
                        if (origValue > 0)
                        {
                            uint value = mainM[iDiscrParam / args.discrParam_divForActualZoom, iEnergy / args.energy_divForActualZoom] += origValue;

                            if (value > maxMainM)
                            {
                                maxMainM = value;
                            }
                        }

                        //shadowMatrix
                        if (args.shadowMatrixDrawingStyle != GraphControl.ShadowMatrixDrawingStyle.hide)
                        {
                            origValue = shadowMatrix[args.positionDiscrParam + iDiscrParam, args.positionEnergy + iEnergy];
                            if (origValue > 0)
                            {
                                uint value = shadowM[iDiscrParam / args.discrParam_divForActualZoom, iEnergy / args.energy_divForActualZoom] += origValue;

                                if (value > maxShadowM)
                                {
                                    maxShadowM = value;
                                }
                            }
                        }
                    }
                }
            }

            uint maxMainMSaturated = maxMainM * (uint)args.colorSaturation / 1000;
            this.maxNumSamplesInGraph = maxMainMSaturated;
            float convMainM = (this.colorScale.Length / (float)maxMainMSaturated);
            float convShadowM = (128 / (float)maxShadowM);

            Bitmap b = null;
            if (args.redrawOnlySectionPreview)
            {
                b = (Bitmap)bitmapPrevious.Clone();
            }
            else
            {
                b = (Bitmap)bitmapWhite.Clone();
            }

            


            //draw 3D graph - without separation lines, section preview etc.
            BitmapData data = b.LockBits(new Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            int stride = data.Stride;
            unsafe
            {
                byte* ptr = (byte*)data.Scan0;

                // Go through the draw area and set the pixels as they should be
                if (!args.redrawOnlySectionPreview)
                {
                    for (int y = 0; y < 512; y++)
                    {
                        for (int x = 0; x < 512; x++)
                        {
                            byte shadowM0 = 0;
                            byte shadowM1 = 0;
                            byte shadowM2 = 0;

                            //shadowMatrix
                            if (shadowM[x, y] > 0 && args.shadowMatrixDrawingStyle != GraphControl.ShadowMatrixDrawingStyle.hide)
                            {
                                int point = (int)(convShadowM * shadowM[x, y]);
                                point = Math.Max(0, Math.Min(255, point));

                                //draw in gray scale
                                point = point / 2 + 128;
                                shadowM0 = ptr[(x * 3) + y * stride] = (byte)point;
                                shadowM1 = ptr[(x * 3) + y * stride + 1] = (byte)point;
                                shadowM2 = ptr[(x * 3) + y * stride + 2] = (byte)point;
                            }

                            //mainMatrix
                            if (mainM[x, y] > 0)
                            {
                                int point = (int)(convMainM * mainM[x, y]);
                                point = Math.Max(0, Math.Min(255, point));

                                //draw in color
                                if (mainM[x, y] <= maxMainMSaturated)
                                {
                                    //normal color
                                    ptr[(x * 3) + y * stride] = colorScale[point].B;
                                    ptr[(x * 3) + y * stride + 1] = colorScale[point].G;
                                    ptr[(x * 3) + y * stride + 2] = colorScale[point].R;
                                }
                                else
                                {
                                    //saturated value will be white
                                    ptr[(x * 3) + y * stride] = 255;
                                    ptr[(x * 3) + y * stride + 1] = 255;
                                    ptr[(x * 3) + y * stride + 2] = 255;
                                }
                            }

                            //redraw if shadow matrix should be above main matrix
                            if (args.shadowMatrixDrawingStyle == GraphControl.ShadowMatrixDrawingStyle.above && shadowM[x, y] > 0)
                            {
                                ptr[(x * 3) + y * stride] = shadowM0;
                                ptr[(x * 3) + y * stride + 1] = shadowM1;
                                ptr[(x * 3) + y * stride + 2] = shadowM2;
                            }
                        }
                    }
                }
            }
            b.UnlockBits(data);

            //save this bitmap
            if (!args.redrawOnlySectionPreview)
            {
                this.bitmapPrevious = (Bitmap)b.Clone();
            }

            //draw section preview, separation line etc.
            data = b.LockBits(new Rectangle(0, 0, b.Width, b.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            stride = data.Stride;
            unsafe
            {
                byte* ptr = (byte*)data.Scan0;

                //draw top separation line
                int topSeparationLine = this.ComputeEnergyCoords(photonNeutronSeparation.EnergyOfTopLine, args);
                if (topSeparationLine < 512 && topSeparationLine >= 0)
                {
                    for (int x = 0; x < 512; x++)
                    {
                        ptr[(x * 3) + topSeparationLine * stride] = 255;
                        ptr[(x * 3) + topSeparationLine * stride + 1] = 0;
                        ptr[(x * 3) + topSeparationLine * stride + 2] = 0;
                    }
                }

                //draw bottom separation line
                int bottomSeparationLine = this.ComputeEnergyCoords(photonNeutronSeparation.EnergyOfBottomLine, args);
                if (bottomSeparationLine < 512 && bottomSeparationLine >= 0)
                {
                    for (int x = 0; x < 512; x++)
                    {
                        ptr[(x * 3) + bottomSeparationLine * stride] = 255;
                        ptr[(x * 3) + bottomSeparationLine * stride + 1] = 0;
                        ptr[(x * 3) + bottomSeparationLine * stride + 2] = 0;
                    }
                }

                //draw vertical separation line
                for (int y = 0; y < 512; y++)
                {
                    int i = y * args.energy_divForActualZoom / args.energy_zoomBitmap + args.positionEnergy;
                    int discrParam = photonNeutronSeparation.SeparationLine[i];
                    if (discrParam >= args.positionDiscrParam)
                    {
                        int x = (discrParam - args.positionDiscrParam) / args.discrParam_divForActualZoom;
                        if (x < 512 && x >= 0)
                        {
                            ptr[(x * 3) + y * stride] = 255;
                            ptr[(x * 3) + y * stride + 1] = 0;
                            ptr[(x * 3) + y * stride + 2] = 0;
                        }
                    } 
                }

                //draw section preview line
                int sectionPreviewR = this.ComputeEnergyCoords(mainMatrix.sectionPreviewR, args);
                if (sectionPreviewR < 512 && sectionPreviewR >= 0)
                {
                    for (int x = 0; x < 512; x++)
                    {
                        ptr[(x * 3) + sectionPreviewR * stride] = 0;
                        ptr[(x * 3) + sectionPreviewR * stride + 1] = 0;
                        ptr[(x * 3) + sectionPreviewR * stride + 2] = 255;
                    }
                }
            }
            b.UnlockBits(data);



            return b.Clone(new Rectangle(0, 0, b.Width / args.discrParam_zoomBitmap, b.Height / args.energy_zoomBitmap), System.Drawing.Imaging.PixelFormat.Format24bppRgb);
        }
    }
}
