﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Dcapture
{
    public class MatrixCSV
    {
        public MatrixCSV(PhotonNeutronSeparation photonNeutronSeparation)
        {
            this.photonNeutronSeparation = photonNeutronSeparation;

            this.matrix = new uint[(int)Math.Pow(2, discrParamSizeInBits), (int)Math.Pow(2, energySizeInBits)];
            this.Clear();
        }

        PhotonNeutronSeparation photonNeutronSeparation = null;

        /// <summary>
        /// 3D matrix where the data are stored
        /// </summary>
        private uint[,] matrix = null;

		public uint this[int discrParam, int energy]
	    {
            get { return matrix[discrParam, energy]; }
	    }

        public int sectionPreviewR = -200;

        /// <summary>
        /// the real width will be 2^widthInBits
        /// </summary>
        private static int energySizeInBits = 12;  //4096
        public static int EnergySizeInBits
        {
            get { return energySizeInBits; }
        }
        public static int EnergySizeMatrix
        {
            get { return (int)Math.Pow(2, energySizeInBits); }
        }
        
        /// <summary>
        /// the real width will be 2^heightInBits
        /// </summary>
        private static int discrParamSizeInBits = 12;   //4096
        public static int DiscrParamSizeInBits
        {
            get { return discrParamSizeInBits; }
        }
        public static int DiscrParamSizeMatrix
        {
            get { return (int)Math.Pow(2, discrParamSizeInBits); }
        }

        public void AddFromCSV(string fileName, bool swapXandYaxe)
        {
            StreamReader sr = new StreamReader(fileName);

            //file header
            string[] line1 = sr.ReadLine().Split('=');
            string[] line2 = sr.ReadLine().Split('=');
            string line3 = sr.ReadLine();

            //check file header
            if (line1[0] != "XRANGE" || line2[0] != "YRANGE" || line3 != "[DATA]")
            {
                throw new Exception("Bad text on line 1 or 2 or 3. Expected is XRANGE or YRANGE or [DATA]");
            }

            //matrix range
            int Xrange = Convert.ToInt32(line1[1]);
            int Yrange = Convert.ToInt32(line2[1]);
            if ( (swapXandYaxe && (Yrange > matrix.GetLength(0) || Xrange > matrix.GetLength(1))) || (!swapXandYaxe && (Yrange > matrix.GetLength(1) || Xrange > matrix.GetLength(0))) )
            {
                throw new Exception("Internal matrix is smaller than matrix in CSV file.");
            }
        
            //file body
            while (!sr.EndOfStream)
            {
                char[] separ = { ' ' };
                string[] line = sr.ReadLine().Split(separ, StringSplitOptions.RemoveEmptyEntries);
                int x = Convert.ToInt32(line[0]);
                int y = Convert.ToInt32(line[1]);
                uint count = Convert.ToUInt32(line[2]);

                if (count > 0)
                {
                    if (swapXandYaxe)
                    {
                        int p = x;
                        x = y;
                        y = p;
                    }

                    matrix[x, y] += count;
                    totalSamples += count;
                    furthermostSampleDiscrParam = Math.Max(x, furthermostSampleDiscrParam);
                    furthermostSampleEnergy = Math.Max(y, furthermostSampleEnergy);

                    this.AddParticlesToSum(y, x, count);
                }
            }

            sr.Close();

            isChanged = true;
        }

        public void SaveToCSV(string fileName)
        {
            StreamWriter sw = new StreamWriter(fileName);

            //file header
            sw.WriteLine("XRANGE=" + (EnergySizeMatrix - 1).ToString());
            sw.WriteLine("YRANGE=" + (DiscrParamSizeMatrix - 1).ToString());
            sw.WriteLine("[DATA]");
            
            
            //file body
            for (int j = 0; j < EnergySizeMatrix; j++)
            {
                for (int i = 0; i < DiscrParamSizeMatrix; i++)
                {
                    if (matrix[i, j] > 0)
                    {
                        sw.WriteLine(j.ToString().PadRight(7) + i.ToString().PadRight(7) + matrix[i, j].ToString());
                    }
                }
            }

            sw.Close();
        }

        

        public void Clear()
        {
            for (int i = 0; i < matrix.GetLength(0); i++)
                for (int j = 0; j < matrix.GetLength(1); j++)
                {
                    matrix[i, j] = 0;
                }

            //position of furthermost sample
            furthermostSampleEnergy = 100;          //minimum is because of drawing
            furthermostSampleDiscrParam = 100;      //minimum is because of drawing

            totalSamples = 0;
            this.ClearSumOfParticles();
        }

        private ulong totalSamples;
        public ulong TotalSamples
        {
            get
            {
                return totalSamples;
            }
        }

        public void AddMatrix(MatrixCSV matrix)
        {
            if (this.matrix.GetLength(0) != matrix.matrix.GetLength(0) || this.matrix.GetLength(1) != matrix.matrix.GetLength(1))
            {
                throw new Exception("Matrix dimensions are not the same.");
            }
            
            for (int i = 0; i < this.matrix.GetLength(0); i++)
                for (int j = 0; j < this.matrix.GetLength(1); j++)
                {
                    uint count = matrix.matrix[i, j];
                    
                    if (count > 0)
                    {
                       this.matrix[i, j] += count;
                       this.AddParticlesToSum(j, i, count);

                       totalSamples += count;
                       furthermostSampleDiscrParam = Math.Max(i, furthermostSampleDiscrParam);
                       furthermostSampleEnergy = Math.Max(j, furthermostSampleEnergy);
                    }
                }
        }
        
        /// <summary>
        /// 
        /// </summary>
        /// <param name="energy"></param>
        /// <param name="discriParam"></param>
        /// <returns>return 1 if the sample is out of the range of the matrix, 0 otherwise</returns>
        public byte AddSample(uint energy, uint discriParam)
        {
            byte outOfSample = 0;
           
            if (discriParam >= matrix.GetLength(0))
            {
                outOfSample = 1;
                discriParam = (uint)(matrix.GetLength(0) - 1);
            }
            if (energy >= matrix.GetLength(1))
            {
                outOfSample = 1;
                energy = (uint)(matrix.GetLength(1) - 1);
            }
            
            matrix[discriParam, energy]++;
            this.AddParticlesToSum((int)energy, (int)discriParam, (uint)1);
            isChanged = true;
            totalSamples++;

            //actualize the position of the furthermost sample
            if (furthermostSampleDiscrParam < discriParam)
            {
                furthermostSampleDiscrParam = (int)discriParam;
            }
            if (furthermostSampleEnergy < energy)
            {
                furthermostSampleEnergy = (int)energy;
            }

            return outOfSample;
        }

        public void GetSectionThroughtDiscrParam()
        {
            throw new System.NotImplementedException();
        }

        public void GetTotalSpectrum()
        {
            throw new System.NotImplementedException();
        }

        public void GetPhotonSpectrum()
        {
            throw new System.NotImplementedException();
        }

        public void GetNeutronSpectrum()
        {
            throw new System.NotImplementedException();
        }

        private bool isChanged = false;
        public bool IsChanged
        {
            get { return isChanged; }
        }

        public void ClearIsChanged()
        {
            isChanged = false;
        }

        private int furthermostSampleEnergy = 0;
        public int FurthermostSampleEnergy
        {
            get { return furthermostSampleEnergy; }
        }

        private int furthermostSampleDiscrParam = 0;
        public int FurthermostSampleDiscrParam
        {
            get { return furthermostSampleDiscrParam; }
        }

        ulong particlesInA = 0;
        public ulong ParticlesInA
        {
            get { return particlesInA; }
        }

        ulong particlesInB = 0;
        public ulong ParticlesInB
        {
            get { return particlesInB; }
        }

        ulong particlesInC = 0;
        public ulong ParticlesInC
        {
            get { return particlesInC; }
        }

        ulong particlesInD = 0;
        public ulong ParticlesInD
        {
            get { return particlesInD; }
        }

        ulong particlesInE = 0;
        public ulong ParticlesInE
        {
            get { return particlesInE; }
        }

        ulong particlesInF = 0;
        public ulong ParticlesInF
        {
            get { return particlesInF; }
        }


        public void ClearSumOfParticles()
        { 
            particlesInA = 0;
            particlesInB = 0;
            particlesInC = 0;
            particlesInD = 0;
            particlesInE = 0;
            particlesInF = 0;
        }

        public void ComputeSumOfParticles()
        {
            ClearSumOfParticles();

            int endE = MatrixCSV.EnergySizeMatrix;
            int endDP = MatrixCSV.DiscrParamSizeMatrix;

            for (int dp = 0; dp < endDP; dp++)
            {
                for (int e = 0; e < endE; e++)
                {
                    uint value = matrix[dp, e];
                    if (value > 0)
                    {
                        AddParticlesToSum(e, dp, value);
                    }
                }
            }
        }

        public void AddParticlesToSum(int energy, int discrParam, uint count)
        {   
            if (discrParam < photonNeutronSeparation.SeparationLine[energy])
            {
                //photons (A, C, E)
                if (energy < photonNeutronSeparation.EnergyOfTopLine)
                {
                    particlesInA += count;
                }
                else if (energy > photonNeutronSeparation.EnergyOfBottomLine)
                {
                    particlesInE += count;
                }
                else
                {
                    particlesInC += count;
                }
            }
            else
            { 
                //neutrons (B, D, F)
                if (energy < photonNeutronSeparation.EnergyOfTopLine)
                {
                    particlesInB += count;
                }
                else if (energy > photonNeutronSeparation.EnergyOfBottomLine)
                {
                    particlesInF += count;
                }
                else
                {
                    particlesInD += count;
                }
            }
        }
    }
}
