﻿using System;
using System.Collections.Generic;
using System.Net.Sockets;
using System.Net;

namespace Dcapture
{
    public class UDPreceiver
    {
        private MatrixCSV destinationMatrix;

        System.ComponentModel.BackgroundWorker backgroundWorker = new System.ComponentModel.BackgroundWorker();
    
        public UDPreceiver(MatrixCSV destinationMatrix)
        {
            this.destinationMatrix = destinationMatrix;
            backgroundWorker.WorkerReportsProgress = true;
            backgroundWorker.DoWork += new System.ComponentModel.DoWorkEventHandler(backgroundWorker_DoWork);
            backgroundWorker.ProgressChanged += new System.ComponentModel.ProgressChangedEventHandler(backgroundWorker_ProgressChanged);
            backgroundWorker.WorkerSupportsCancellation = true;
            backgroundWorker.RunWorkerCompleted += new System.ComponentModel.RunWorkerCompletedEventHandler(backgroundWorker_RunWorkerCompleted);
        }

        public delegate void UDPreceiverErrorHandler(string message);
        public event UDPreceiverErrorHandler UDPreceiverError;

        /// <summary>
        /// exception occurs while udp receiving
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void backgroundWorker_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
        {
            if (e.Cancelled)
            {
                if (UDPreceiverError != null)
                {
                    UDPreceiverError(backgroundWorkerErrorMsg);
                }
            }

            //UDP receiver is stopped
            if (UDPreceiverMessage != null)
            {
                UDPreceiverMessage(UDPreceiverMessageEnum.stopped);
            }
        }

        /// <summary>
        /// called when packet is received or state is changed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void backgroundWorker_ProgressChanged(object sender, System.ComponentModel.ProgressChangedEventArgs e)
        {
            //message that inform that new packet was received
            if (UDPreceiverMessage != null)
            {
                UDPreceiverMessage((UDPreceiverMessageEnum)e.UserState);
            }
        }

        public delegate void UPDreceiverMessageHandler(UDPreceiverMessageEnum message);
        public event UPDreceiverMessageHandler UDPreceiverMessage;

        public enum UDPreceiverMessageEnum
        { 
            stopped,
            running,
            newPacket,
            noChange
        }

        private bool stopRequest = false;

        private DateTime startTime = DateTime.Now;
        public DateTime StartTime
        {
            get
            {
                return startTime;
            }
        }

        private DateTime endTime = DateTime.Now;
        public DateTime EndTime
        {
            get
            {
                return endTime;
            }
        }

        private uint catchedPackets = 0;
        public uint CatchedPackets
        {
            get
            {
                return catchedPackets;
            }
        }

        private uint catchedSamples = 0;
        public uint CatchedSamples
        {
            get
            {
                return catchedSamples;
            }
        }

        private uint lostPackets = 0;
        public uint LostPackets
        {
            get
            {
                return lostPackets;
            }
        }

        private uint samplesOutOfMatrix = 0;
        public uint SamplesOutOfMatrix
        {
            get
            {
                return samplesOutOfMatrix;
            }
        }

        public void Start(int port)
        {
            if (backgroundWorker.IsBusy)
            {
                throw new Exception("UDP receiver is starting or running");
            }
            else
            {
                stopRequest = false;
                backgroundWorker.RunWorkerAsync(port);
            }
        }

        public void Stop()
        {
            stopRequest = true;
        }

        string backgroundWorkerErrorMsg = "";


        void backgroundWorker_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            UdpClient udpClient = null;
            IPEndPoint RemoteIpEndPoint = null;

            //open UDP
            try
            {
                int port = (int)e.Argument;
                //IPEndPoint object will allow us to read datagrams sent from any source.
                RemoteIpEndPoint = new IPEndPoint(IPAddress.Any, port);

                // This constructor arbitrarily assigns the local port number.
                udpClient = new UdpClient(RemoteIpEndPoint);
                udpClient.Client.ReceiveTimeout = 1000;
            }
            catch (Exception exc)
            {
                backgroundWorkerErrorMsg = "Error while port reservation and UDP opening\n\n" + exc.Message;
                e.Cancel = true;
                return;
            }

            uint previousUDPpacketNum = 0;
            bool firstRun = true;
            samplesOutOfMatrix = 0;
            lostPackets = 0;
            catchedPackets = 0;
            catchedSamples = 0;

            startTime = DateTime.Now;
            backgroundWorker.ReportProgress(1, UDPreceiverMessageEnum.running);
            DateTime previousTimeOfPacket = DateTime.Now;

            while (!stopRequest)
            {
                try
                {
                    Byte[] rec = udpClient.Receive(ref RemoteIpEndPoint);
                    
                    //packet header
                    uint UDPpacketNum = BitConverter.ToUInt32(rec, 0);
                    ushort numPulsesInPacket = BitConverter.ToUInt16(rec, 4);
                    uint pulseNum = BitConverter.ToUInt32(rec, 6);
                    ulong pulseTime = BitConverter.ToUInt64(rec, 10);
                    //Console.Write(UDPpacketNum.ToString("D6") + " " + numPulsesInPacket.ToString("D4") + " " + pulseNum.ToString("D6") + " " + pulseTime.ToString("D10") + " ");
                    
                    //packet body
                    for (int i = 0; i < numPulsesInPacket; i++)
                    {
                        uint energy = BitConverter.ToUInt16(rec, 18 + i * 4);
                        uint discrParam = BitConverter.ToUInt16(rec, 18 + i * 4 + 2);
                        //Console.Write(" | " + energy.ToString() + " " + discrParam.ToString());

                        samplesOutOfMatrix += destinationMatrix.AddSample(energy, discrParam);
                        
                    }
                    //Console.WriteLine();

                    //lost packets counter
                    if (UDPpacketNum != previousUDPpacketNum + 1 && !firstRun)
                    {
                        uint lost = UDPpacketNum - (previousUDPpacketNum + 1);
                        this.lostPackets += lost;
                        //Console.WriteLine("lost packets: " + lost.ToString("D6") + "total lost: " + lostPackets.ToString("D6"));
                    }

                    //cached packets counter
                    this.catchedPackets += 1;
                    this.catchedSamples += numPulsesInPacket;

                    firstRun = false;
                    previousUDPpacketNum = UDPpacketNum;
                    if ((DateTime.Now - previousTimeOfPacket).TotalSeconds >= 1)
                    {
                        backgroundWorker.ReportProgress(1, UDPreceiverMessageEnum.newPacket);
                        previousTimeOfPacket = DateTime.Now;
                    }
                }
                catch (System.Net.Sockets.SocketException exc)
                {
                    if (exc.ErrorCode == 10060)
                    {
                        //it is OK, it is time out -> report no change
                        backgroundWorker.ReportProgress(1, UDPreceiverMessageEnum.noChange);
                    }
                    else
                    {
                        backgroundWorkerErrorMsg = "Error in UDP receiver.\n\n" + exc.Message;
                        udpClient.Close();
                        e.Cancel = true;
                        return;
                    }
                }
                catch (Exception exc)
                {
                    backgroundWorkerErrorMsg = "Error in UDP receiver.\n\n" + exc.Message;
                    udpClient.Close();
                    e.Cancel = true;
                    return;
                }
            }
            udpClient.Close();

            endTime = DateTime.Now;
        }
    }
}
